\\ Copyright 2016 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  If not, see <http://www.gnu.org/licenses/>.


\\ This is an example of using lindep() to express a linear recurrence in
\\ terms of some desired set of terms.
\\
\\ foo() is a formula in n and fibonacci(n)*fibonacci(n+/-1) and the aim is
\\ to express it in terms of fibonacci(n)^2, fibonacci(n+1)^2, etc.  Some
\\ fibonacci identities and manipulations of the formula can do this, but a
\\ mechanical way is to use lindep on generating functions.
\\
\\ Vector v is generating functions ("t_RFRAC"s) for foo and the desired
\\ terms.  lindep() gives factors on foo and the terms which sum to 0,
\\ thereby expressing foo in terms of the others.  lindep() won't act on
\\ t_RFRACs (as of gp 2.7.x) so it's necessary to multiply up by a common
\\ denominator to give plain polynomials.
\\
\\ In general, lindep() will give an empty column []~ if there's no
\\ relationship.  Usually that means more powers or combinations are needed.
\\ Terms not in fact needed come out with a 0 factor.
\\
\\ It also works to use just Pol() instead of recurrence_guess_values_to_gf(), to find a 
\\

default(recover,0);
read("recurrence-guess.gp");

foo(n) = (n*fibonacci(n) + fibonacci(n))*(n*fibonacci(n+1) + fibonacci(n-1));
v= {[
     recurrence_guess_values_to_gf(vector(50,n,n--; foo(n))),

     recurrence_guess_values_to_gf(vector(50,n,n--; n^2*fibonacci(n+2)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; n*fibonacci(n+2)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; fibonacci(n+2)^2)),

     recurrence_guess_values_to_gf(vector(50,n,n--; n^2*fibonacci(n+1)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; n*fibonacci(n+1)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; fibonacci(n+1)^2)),

     recurrence_guess_values_to_gf(vector(50,n,n--; n^2*fibonacci(n)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; n*fibonacci(n)^2)),
     recurrence_guess_values_to_gf(vector(50,n,n--; fibonacci(n)^2))
   ]};

lindep(v*lcm(apply(denominator,v)))
\\ == [-2, 1, 2, 1, -1, -2, -1, -1, -4, -3]~

\\ thus giving
foo_by_squares(n) =
{
  1/2*(  (n^2+2*n+1) * fibonacci(n+2)^2
       - (n^2+2*n+1) * fibonacci(n+1)^2 
       - (n^2+4*n+3) * fibonacci(n)^2
      );
}
vector(100,n,foo(n)) == vector(100,n,foo_by_squares(n))
