\\ Copyright 2016, 2018, 2019, 2020, 2021, 2022, 2024 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  If not, see <http://www.gnu.org/licenses/>.

default(recover,0);
default(strictargs,1);
iferr(read("recurrence-guess.gp"),e,read("../recurrence-guess.gp"));


\\------------------------------------------------------------------------------
\\ Eventually All 0s

recurrence_guess([0,0,0,0,0,0,0,0,0])
\\ recurrence_guess([1,2,3,0,0,0,0,0,0,0,0,0,0,0,0])
quit



\\------------------------------------------------------------------------------
\\ Coefficients not integers

a(n) = prod(k=0,6, (n+k)\7);
recurrence_guess(vector(500,n,n--; if(n==0,0,a(n))));
quit


\\------------------------------------------------------------------------------

recurrence_guess(OEIS_samples("A236576"))
quit

\\------------------------------------------------------------------------------

\\ values=vector(10,n,2^n+3^n+5^n);
\\ 
\\ M=matrix(3,3,i,j,values[i+j-1]);
\\ C=vectorv(3,i,values[i+3]);
\\ matsolve(M,C)
\\ 
\\ L=matrix(4,4,i,j,values[i+j-1]);
\\ lindep(L)
\\ 
\\ recurrence_guess(values)

default(parisizemax,100*10^6)
len=100;
\\ linear recurrence of order "len"
values=vector(3*len,n, sum(i=1,len, i^n));

\\ fast
L=matrix(len+1,len+1,i,j,values[i+j-1]);
gettime(); lindep(L); gettime()

\\ slower
M=matrix(len,len,i,j,values[i+j-1]);
C=vectorv(len,i,values[i+len]);
gettime(); matsolve(M,C); gettime()

\\ struggles on big numbers?
gettime(); bestapprPade(Ser(values)); gettime()

print();
lindep(L)
matsolve(M,C)
bestapprPade(Ser(values))

quit;



\\------------------------------------------------------------------------------
recurrence_guess(Vec(1/(1 - x^50) + O(x^500)))
quit


\\------------------------------------------------------------------------------

f(n) =
{
  if(n==0,10,
     n==1,20,
     n==2,30,
     -f(n-1) - 3*f(n-2) + f(n-3));
}
recurrence_guess(vector(20,n,n--; if(n==0,999, f(n))));
quit

\\------------------------------------------------------------------------------

{
  \\ irreducible
  recurrence_guess([0, 1, 1, 2, 4, 10, 22, 45, 93, 197, 420, 890, 1878, 3964,              8380, 17724, 37474, 79209, 167426, 353927, 748202, 1581668,              3343519, 7067929, 14941121, 31584586, 66767743, 141142461,              298365531, 630724607, 1333309486, 2818526581, 5958175175,              12595180550, 26625362976, 56284223827]);

  recurrence_guess(OEIS_samples("A246178"));
  recurrence_guess(Vec(1/87 * (3 - 6*'x + 4*'x^2)/(1 - 'x)^5/(1-2*x)/(1-3*x)^2 + O(x^100)));
  quit;
}
{
  recurrence_guess(vector(10,n, 'a^n + n*'b + 'c + 'a));
  quit;
}
\\ numerator of factorized gf
recurrence_guess(OEIS_samples("A246178"))
quit


\\------------------------------------------------------------------------------

M = x^3 - 17*x^2 - 13*x - 11;
gettime(); for(reps=1,1000, x^1000 % M); gettime()
gettime(); for(reps=1,1000, Mod(x,M)^1000); gettime()
quit


\\------------------------------------------------------------------------------

\\ OEIS A077949 1/(1-x-2*x^3)
\\ signature (1,0,2)
\\
\\ v[n-3]* [2, 0, 1] *v[n-1]  = v[n]
\\ v[n] =  v[n-1]* [1, 0, 2] *v[n-3]

v=Vec(Ser((1-x)/(1 - x - 2*x^3)))

lindep([v[1..5],v[2..6],v[3..7],v[4..8],v[5..9]])

v=Vec(Ser(1/(1-x)/(1 - x - 2*x^3)))
recurrence_guess(v);

v=Vec(Ser(1/polrecip(x^3 - 5*x^2 + x - 1)))
recurrence_guess(v);

lindep([1/(1-x)/(1 - x - 2*x^3), 1/(1 - x - 2*x^3)])
lindep([(1-x)/(1 - x - 2*x^3), (1-2*x)/(1 - x - 2*x^3), (2-x)/(1 - x - 2*x^3)] \
       * (1 - x - 2*x^3))
quit


\\------------------------------------------------------------------------------

recurrence_guess_INTERNAL_print_gf((33 - 151*'x + 264*'x^2 - 210*'x^3 + 65*'x^4)/(1 - 'x)^5);
recurrence_guess_INTERNAL_print_gf(1/87 * (3 - 6*'x + 4*'x^2)/(1 - 'x)^5);
quit

nums_exp=[0];
i=1;
d=1;
num=(-4*x^2 + 6*x - 3)/(-x^2 + 2*x - 1);
while(num%d==0, nums_exp[i]++; num/=d; if(d==1,break));
quit

recurrence_guess_INTERNAL_print_gf((3 - 6*'x + 4*'x^2)/(1 - 'x)^5);
quit


\\------------------------------------------------------------------------------

gf_terms(g,n) = {
  if(g==0,return(vector(n)));
  my(x = variable(g),
     zeros = min(n,valuation(g,x)),
     v = Vec(g + O(x^n)));
  if(zeros>=0, concat(vector(zeros,i,0), v),
               v[-zeros+1 .. #v]);
}

g=1/(1 - x - 2*x^3);
v=gf_terms(g,20)[3..20]
len=5;
m = recurrence_guess_INTERNAL_values_to_matrix(v,len);
a = recurrence_guess_INTERNAL_values_to_column(v,len);
print("m="m);
print("a="a);
print("lindep");
lindep(Vec(m))
print("solve");
s=matsolve(m,a)
m*s
quit

\\------------------------------------------------------------------------------

vector_modulo(v,i) = v[(i% #v)+1];
phi = quadgen(5);    \\ (1 + sqrt(5))/2  D=5

recurrence_guess(vector(50,n,n--; phi*fibonacci(n)));
quit

version()
v=vector(10,k,k--; 1 + 2^k*'y + 3^k*'y^2);
print("v = "v);
f = recurrence_guess_values_to_factors(v);
print("f = "f);
g = recurrence_guess_factors_and_values_to_gf(f,v);
print("g = "g);
recurrence_guess(v)
quit

recurrence_guess(recurrence_guess_INTERNAL_gf_terms(9+99*x+x^2*(1+x)/(1-2*x)^3/(1-x), 50));
quit

\\ [1, 0, 0, 0, 8, 0, 0, 0, 48, 0, 0, 0, 256, 0, 0, 0, 1280, 0, 0, 0]
print(recurrence_guess_INTERNAL_gf_terms(1/(1-4*x^4)^2, 20));
{
  print(vector(20,n,n--;
         vector_modulo([1/4, 0, 0, 0],n) * n * 2^floor(n/2)
         + vector_modulo([1, 0, 0, 0],n) * 2^floor(n/2)))
}
recurrence_guess(recurrence_guess_INTERNAL_gf_terms(-(1+I)*x/(1-4*x^4)^2, 50));
quit

recurrence_guess(recurrence_guess_INTERNAL_gf_terms \
             ((82*x + 28)/(900*x^4 + 1800*x^3 + 1800*x^2 + 900*x + 225), 50));
quit
recurrence_guess(recurrence_guess_INTERNAL_gf_terms \
             ((-54*x - 28)/(450*x^2 + 450*x + 225), 50));
{
  recurrence_guess(vector(50,n,n--;
    vector_modulo([ 41/225, -28/225, -13/225,  6/25,
                   -41/225,  28/225,  13/225, -6/25 ],n)
    * n * 2^floor(n/2)));
}
quit

recurrence_guess(vector(50,n,n--;   4/25 * n * 2^n));
recurrence_guess(vector(50,n,n--;  -18/125 * 2^n));

recurrence_guess( [ 0, 0, 0, 2, 4, 30, 64, 106, 284, 622, 1528, 3506, 7124, 15702, 34160, 73786, 160076, 337374, 714856, 1518018, 3201988, 6754534, 14167136 ] );
quit

\\------------------------------------------------------------------------------

\\ pol_max_abs_root(p) = vecmax(apply(abs,Vec(polroots(p*1.0))));
\\ cmp_pol_max_denominator_root(x,y) = lex(pol_max_abs_root(denominator(x)), \
\\                                         pol_max_abs_root(denominator(y)));

\\ \\ 6/3==2
\\ \\ 5/3==2
\\
\\ recurrence_guess__INTERNAL_vector_is_all_equal(v) = \
\\   sum(i=2,#v,v[i-1]!=v[i]) == 0;
\\
\\ recurrence_guess__INTERNAL_vector_split(v,n) = \
\\   vector(ceil(#v/n),i, v[(1+(i-1)*n) .. min(#v,i*n)]);
\\ recurrence_guess__INTERNAL_vector_split([11,12,13,14,15],3) == [[11,12,13],[14,15]]
\\ recurrence_guess__INTERNAL_vector_split([11,12,13,14,15,16],3) == [[11,12,13],[14,15,16]]
\\
\\      my(vs = recurrence_guess__INTERNAL_vector_split(v,ds[i]));
\\      if(recurrence_guess__INTERNAL_vector_is_all_equal(vs),
\\         return(vs[1])));
\\
\\   while(m--,
\\         fordiv(#v,d,
\\               if(d<#v,
\\                  )));
\\   v;


     \\ my(base_pos=0);
     \\ for(j=1,#base_list,
     \\    if(base_list[j]==base && degree_list[j]==degree,
     \\       base_pos=j;
     \\       break()));
     \\ if(!base_pos,
     \\    listput(base_list, base);
     \\    listput(degree_list, degree);
     \\    listput(periodic_list, [0]);
     \\    base_pos=#base_list);
     \\
     \\ periodic = vector(#t,i, t[i]/base^floor((i-1)/degree));
     \\ periodic_list[base_pos]
     \\ = recurrence_guess__INTERNAL_periodic_vector_reduce
     \\  (recurrence_guess__INTERNAL_periodic_vector_sum
     \\   (periodic_list[base_pos], periodic));

\\ base_list = List([]),
\\      periodic_list = List([]),
\\      degree_list = List([]),


\\-----------------------------------------------------------------------------
\\ Ascending Output

\\ Possible code trimming an O(x^999) from the end.
\\ It puts parens on complex 3+I though like "(3 + I)".
  \\ my(str="");
  \\ if(want_plus,
  \\    my(v=Vecrev(p));
  \\    for(i=1,#v,
  \\       if(v[i],
  \\          if(v[i]<0,
  \\             str=concat(str,"- ");
  \\             v[i] = - v[i];
  \\             want_plus=0);
  \\          break()));
  \\    p=Polrev(v));
  \\ if(want_plus,
  \\    str="+ ");
  \\ if(p==0,return(0));
  \\ my(o = O(variable(x)^(poldegree(p)+100)),
  \\    vstr=Vecsmall(Str(p+o)),
  \\    vstrend=Vecsmall(Str(" + "o)));
  \\ if(vstr[(#vstr-#vstrend+1) .. #vstr] != vstrend,
  \\    error("Oops, stringized p didn't end with ",Strchr(vstrend),
  \\          " got ",Strchr(vstr)));
  \\ concat(str,Strchr(vstr[1.. (#vstr-#vstrend)]));


\\-----------------------------------------------------------------------------
vm(v,i) = v[(i% #v)+1];

sqrt2 = quadgen(4*2);
sqrt3i = quadgen(4*-3);
sqrt3 = quadgen(4*3);
w6 = 1/2 + 1/2*sqrt3i;
w3 = -1/2 + 1/2*sqrt3i;
w = sqrt2;

\\ my_vector = vector(10,k, fibonacci(k+2))
\\ my_vector = vector(10,k, 4*2^k)
my_vector = vector(10,k, 3^k + 4*2^k*fibonacci(k));
my_vector = vector(10,k, k!);
my_vector = vector(10,k,k--; 3^k + k*2^k);
my_vector = vector(10,k,k--; 2^k+3^k);
my_vector = vector(10,k,k--; vm([7,5,3,11],k));
my_vector = vector(10,k,k--; vm([7,5,3,11],k)*2^k);
my_vector = vector(10,k,k--; vm([7,5,3],k)*3^k);
my_vector = (vector(50,k,k--; vm([7,5,3,11],k)*2^k + 3^k*I));
my_vector = (vector(50,k,k--; k^2));
my_vector = (vector(50,k,k--; k^2 * 2^k));
my_vector = (vector(50,k,k--; k^3+k^2+k+1));
my_vector = (vector(50,k,k--; k^5+1));
my_vector = vector(10,k,k--; (-2)^k+(-3)^k);
my_vector = recurrence_guess_INTERNAL_gf_terms(1/(1-2*x),20);
my_vector = recurrence_guess_INTERNAL_gf_terms(1/(1-2*x^2),20);
my_vector = vector(10,n,n--; n*2^n+1);
my_vector = [0,I, -1 + 2*I, -2 + I, -3 - 2*I, -2 - 5*I, 3 - 6*I, 10 - 3*I, 13 + 6*I, 6 + 19*I, -13 + 26*I];
my_vector = vector(50,k,k--; 1 + 2^k*'y + 3^k*'y^2);
my_vector = vector(50,k,k--; 1 + 2^k*'a);
my_vector = recurrence_guess_INTERNAL_gf_terms((3+5*x)/(1+x^2)^2, 30);

\\ A027656 = 1/(1-x^2)^2  1, 0, 2, 0, 3, 0, ...
my_vector = recurrence_guess_INTERNAL_gf_terms(1/(1+x^2)^2, 200);
\\ recurrence_guess(my_vector);
\\ quit
my_vector = recurrence_guess_INTERNAL_gf_terms(1/(1-sqrt2*x^2), 10);
my_vector = vector(50,k,k--; vm([0,-1,1,-1, 0,1,-1,1],k)*2^floor(2*k/2));
my_vector = vector(100,k,k--; (k+1)*2^k);
my_vector = vector(100,k,k--; k*2^k);
my_vector = vector(100,k,k--; k^3*2^k);
my_vector = vector(100,k,k--; 2^floor(k/3));
my_vector = vector(30,k,k--; (k+1)*2^floor(k/3));
my_vector = vector(30,k,k--; k*2^floor(k/3));
my_vector = vector(300,k,k--; k^5*2^floor(k/3));
my_vector = vector(10,k,k--; 2^k);
my_vector = vector(30,k,k--; (1+I)^k);
my_vector = vector(30,k,k--; k^2*(1+I)^k);
my_vector = vector(30,k,k--; vm([1,0],k));
my_vector = vector(30,k,k--; k);
my_vector = recurrence_guess_INTERNAL_gf_terms( 1/(4 + 4*x), 100);
my_vector = recurrence_guess_INTERNAL_gf_terms( 1/(1 - x^2)^2, 100);
my_vector = vector(30,k,k--; k^2*vm([1,0],k));
my_vector = recurrence_guess_INTERNAL_gf_terms( 1/(1 - x)^2, 100);
my_vector = recurrence_guess_INTERNAL_gf_terms( 1/(1 + x^2)^3, 100);
my_vector = recurrence_guess_INTERNAL_gf_terms( 1/(1 + 2*x^2)^3, 100);
my_vector = vector(30,k,k--; floor(k/3));
my_vector = vector(30,k,k--; floor(k/3)^2);
my_vector = vector(30,k,k--; (1+I)^k + (1-I)^k);
my_vector = vector(30,k,k--; w6^k + conj(w6)^k);
my_vector = vector(30,k,k--; sqrt2 * (-2)^k);
my_vector = vector(50,k,k--; 'y*k*2^k*vm([1,0,0,0],k));
my_vector = vector(50,k,k--; 'y * k^2 * 2^k * vm([1,0,0,0],k));
my_vector = quadgen(20)*[0, 0, 1, 10, 50, 214, 802, 2802, 9275, 29580, 91668, 277924, 828092, 2433140, 7067885, 20337318, 58054534, 164602410, 463990190, 1301338150, 3633753815];  \\ mucho slow in show powers
{
  print(my_vector);
  recurrence_guess(my_vector)
}

f(n) = (4/5 + 2/5*I) * (1 + I)^n + vm([1/5 - 2/5*I, -1/5 + 2/5*I],n) * I^n;
vector(10,n,n--; f(n))
f(n) = (4/5 + 2/5*I) * (1 + I)^n + vm([1/5*I^n - 2/5*I*I^n, -1/5*I^n + 2/5*I*I^n],n);
vector(10,n,n--; f(n))